from flask import Flask, render_template, request
import networkx as nx
import json
from pklx.settings import SETTINGS
from pklx.parser import load, statements_to_graph
from pklx.utils import bfs
import uuid

app = Flask(__name__)

GRAPH = None
NODES = None
COLORS = {'variable': '#7a7a7a', 'relation': '#dd4b39'}


def format_graph(graph: nx.Graph) -> nx.Graph:
    for node in graph.nodes:
        graph.nodes[node]['id'] = str(uuid.uuid4())
        if graph.nodes[node]['node_type'] != 'variable':
            graph.nodes[node]['color'] = COLORS['relation']
            graph.nodes[node]['label'] = " ".join(node.split()[1:])
        else:
            graph.nodes[node]['color'] = COLORS['variable']
            graph.nodes[node]['label'] = node
        graph.nodes[node]['shape'] = 'dot'
        graph.nodes[node]['borderWidth'] = 0
        graph.nodes[node]['size'] = 10
    for source, destination in graph.edges:
        graph[source][destination]['id'] = source + '-/-' + destination
        graph[source][destination]['from'] = source
        graph[source][destination]['to'] = destination
        graph[source][destination]['width'] = 2
        graph[source][destination]['color'] = COLORS['variable']
    return graph


@app.route('/')
def index():
    return render_template('index.html')

@app.route('/nodes', methods=['GET'])
def nodes():
    return NODES

@app.route('/related', methods=['POST'])
def related():
    data = request.get_json()
    node = data['request']
    graph = bfs(GRAPH, node, stop_condition=lambda node: GRAPH.nodes[node]['node_type'] == 'variable')
    graph = format_graph(graph)
    graph.nodes[node]['color'] = '#00a303'
    data = nx.json_graph.node_link_data(graph)
    return json.dumps(data)


def main():
    global GRAPH
    global NODES
    relations, statements = load(SETTINGS['FOLDER_PATH'])

    if not relations and not statements:
        print(f'No relations or statements found at {SETTINGS["FOLDER_PATH"]}. Please use pklx-set-settings FOLDER_PATH <absolute_path_to_data_folder> to set the path to the data folder.')
        exit(1)

    GRAPH = statements_to_graph(statements)
    node_names = [node for node in GRAPH.nodes if GRAPH.nodes[node]['node_type'] == 'variable']
    NODES = json.dumps([{"id": node, "text": node} for node in node_names])
    app.run()


if __name__ == '__main__':
    main()



import argparse
from .visx.backend import main as main_view
from pklx.parser import load, extract_from_statements
from .settings import SETTINGS
from .settings import set_settings as internal_set_settings


def view():
    main_view()


def collect(variable, file):
    relations, statements = load(SETTINGS['FOLDER_PATH'])
    extracted_statements = extract_from_statements(statements, variable)
    if file is not None:
        statement_text = '\n'.join(map(str, extracted_statements))
        with open(file, 'w') as f:
            f.write(statement_text)
    else:
        for statement in extracted_statements:
            print(statement)


def set_settings(key, value):
    internal_set_settings(key, value)


def cmd_view():
    main_view()


def cmd_collect():
    parser = argparse.ArgumentParser(description='Collect knowledge about a specific variable')
    parser.add_argument('variable', type=str, help='The variable to collect knowledge about')
    parser.add_argument('--file', type=str, help='Save the output into a file')
    args = parser.parse_args()
    collect(args.variable, args.file)


def cmd_set_settings():
    parser = argparse.ArgumentParser(description='Set settings')
    parser.add_argument('key', type=str, help='The key of the setting')
    parser.add_argument('value', type=str, help='The value of the setting')
    args = parser.parse_args()
    internal_set_settings(args.key, args.value)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Train a network.')
    parser.add_argument('--view', action='store_true', help='View the network')
    parser.add_argument('--collect', type=str, help='Collect knowledge about a specific variable')
    parser.add_argument('--file', type=str, help='Save the output into a file')
    parser.add_argument('--settings', nargs='+', type=str, help='Specify settings by passing a list of key-value pairs. Example: --settings FOLDER_PATH /home/user/ DELIMITER /')

    args = parser.parse_args()

    if args.collect is not None:
        relations, statements = load(SETTINGS['FOLDER_PATH'])
        extracted_statements = extract_from_statements(statements, args.collect)
        if args.file is not None:
            statement_text = '\n'.join(map(str, extracted_statements))
            with open(args.file, 'w') as f:
                f.write(statement_text)
        else:
            for statement in extracted_statements:
                print(statement)
    if args.view:
        main_view()
    if args.settings is not None:
        for i in range(0, len(args.settings), 2):
            key = args.settings[i]
            value = args.settings[i + 1]
            set_settings(key, value)



from pklx.parser import load, extract_from_statements
from pklx.settings import SETTINGS

if __name__ == '__main__':
    relations, statements = load(SETTINGS['FOLDER_PATH'])
    extracted_statements = extract_from_statements(statements, 'B')
    for statement in extracted_statements:
        print(statement)



import os
import re
from typing import List, Tuple
from .objects import PKLX, Statement
import networkx as nx
from .settings import SETTINGS


def load(folder_path: str) -> Tuple[List[str], List[str]]:
    statements = []
    relations = []
    file_names = [os.path.join(folder, file) for folder, _, files in os.walk(folder_path) for file in files]
    for file_name in file_names:
        if os.path.split(file_name)[-1] == '.ontology':
            with open(os.path.join(folder_path, file_name), 'r') as file:
                relations = file.readlines()
            relations = parse_relations(relations)
        else:
            with open(os.path.join(folder_path, file_name), 'r') as file:
                lines = file.readlines()
            for line in lines:
                # separate statements from text by using the delimiter
                splitted_line = line.split(SETTINGS['DELIMITER'])
                if len(splitted_line) % 2 == 0:
                    raise Exception(f'Invalid syntax in file: {file_name}, line: {line}')
                else:
                    statements.extend(splitted_line[1::2])
    statements = parse_statements(relations, statements)
    return relations, statements


def parse_relations(relations: List[str]) -> List[str]:
    return [relation.split(SETTINGS['DELIMITER'])[0].strip().replace('\n', '') for relation in relations]


def lexer(relations: List[str], statement: str) -> List[str]:
    split_tokens = [token for token in re.split(r'(\W)', statement) if token.strip()]
    relations = relations.copy()
    # longer relations should be matched first
    relations.sort(key=len, reverse=True)
    # merge as many consecutive tokens into a relation as possible
    relation_merged_tokens = []
    i = 0
    while i < len(split_tokens):
        found = False
        for phrase in relations:
            if " ".join(split_tokens[i:i+len(phrase.split())]) == phrase:
                relation_merged_tokens.append(phrase)
                i += len(phrase.split())
                found = True
                break
        if not found:
            relation_merged_tokens.append(split_tokens[i])
            i += 1

    # merge consecutive alphanumeric tokens that are not relations into one token
    tokens = []
    current_string = ""
    for s in relation_merged_tokens:
        if s.isalnum() and s not in relations:
            current_string += " " + s
        else:
            if current_string:
                tokens.append(current_string.strip())
                current_string = ""
            tokens.append(s)
    if current_string:
        tokens.append(current_string.strip())

    # remove optional markdown syntax
    i = 0
    while i < len(tokens) - 1:
        if tokens[i] == '[' and tokens[i+1] == '[':
            del tokens[i:i+2]
            continue
        if tokens[i] == ']' and tokens[i+1] == ']':
            del tokens[i:i+2]
            continue
        i += 1

    return tokens


def parse_statements(relations: List[str], statements: List[str]) -> List[PKLX]:
    parsed_statements = []
    for statement in statements:
        parsed_statements.append(parse_statement(relations, statement))
    return parsed_statements


def parse_statement(relations: List[str], statement: str) -> PKLX:
    tokens = lexer(relations, statement)
    parsed_statement = PKLX().parse(tokens, relations)
    return parsed_statement


def statements_to_graph(statements: List[PKLX]) -> nx.DiGraph:
    graph = nx.compose_all([statement.to_graph()[1] for statement in statements])
    for node in list(graph.nodes):
        try:
            variable = graph.nodes[node]['variable']
            nx.contracted_nodes(graph, node, variable, copy=False)
        except KeyError:
            pass
    return graph


def extract_from_statements(statements: List[PKLX], variable: str, done: List[str] = None) -> List[PKLX]:
    extracted_statements = []
    statement_variables = []
    if done is None:
        done = []
    if variable in done:
        return []
    for statement in statements:
        if statement.contains(variable):
            extracted_statements.append(statement)
            if type(statement) == Statement:
                statement_variables.extend(statement.variable.name)
    done.append(variable)
    for statement_variable in statement_variables:
        if statement_variable not in done:
            extracted_statements.extend(extract_from_statements(statements, statement_variable, done))
    extracted_statements = list(set(extracted_statements))
    return extracted_statements



import os
import json


SETTINGS = {
    "FOLDER_PATH": '',
    "DELIMITER": '-/'
}


def set_settings(key: str, value: str):
    global FOLDER_PATH
    global DELIMITER
    if key in SETTINGS:
        SETTINGS[key] = value
    else:
        raise Exception(f'Invalid key: {key}')
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'settings.json'), 'w') as file:
        json.dump(SETTINGS, file, indent=4)


with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'settings.json'), 'r') as file:
    SETTINGS = json.load(file)



import networkx as nx
from collections import deque
from typing import Callable


def bfs(graph: nx.Graph, source: str, stop_condition: Callable = None) -> nx.Graph:
    visited = set([source])
    if stop_condition is None:
        stop_condition = lambda node: False
    undirected_graph = graph.to_undirected()
    neighbors = undirected_graph.neighbors(source)
    queue = deque([neighbor for neighbor in neighbors])
    while queue:
        node = queue.pop()
        visited.add(node)
        if not stop_condition(node):
            neighbors = undirected_graph.neighbors(node)
            queue.extendleft([neighbor for neighbor in neighbors if neighbor not in visited])
    return graph.subgraph(visited)



from setuptools import setup

setup(
    name='PKLX',
    version='0.0.3',
    author='PhylomatX',
    author_email='phylomatx@gmail.com',
    description='PKLX',
    install_requires=['networkx==3.1', 'flask==2.3.2'],
)


import networkx as nx
from typing import Tuple
import uuid
from typing import List


class ParsingException(Exception):
    pass


class PKLX():

    def parse(self, tokens: List[str], relations: List[str]) -> 'PKLX':
        try:
            return Statement().parse(tokens, relations)
        except ParsingException:
            pass
        try:
            return Knowledge().parse(tokens, relations)
        except ParsingException:
            pass
        raise ParsingException(f"Could not parse PPLX: {tokens}")


class Statement(PKLX):
    
    def __init__(self, name=None, knowledge=None):
        self.variable = name
        self.knowledge = knowledge

    def parse(self, tokens: List[str], relations: List[str]) -> PKLX:
        if type(tokens) != list or len(tokens) == 0:
            raise ParsingException(f"Could not parse Statement: {tokens}")
        try:
            assert tokens[1] == '='
            self.variable = Name().parse(tokens[0], relations)
            self.knowledge = Knowledge().parse(tokens[2:], relations)
            return self
        except (AssertionError, ParsingException):
            raise ParsingException(f"Could not parse Statement: {tokens}")

    def to_graph(self) -> Tuple[str, nx.DiGraph]:
        anchor, graph = self.knowledge.to_graph()
        graph.nodes[anchor]['variable'] = self.variable.name
        return anchor, graph

    def contains(self, variable: str) -> bool:
        if self.variable.contains(variable):
            return True
        if self.knowledge.contains(variable):
            return True
        return False

    def __repr__(self):
        return f"{self.variable} = {self.knowledge}"


class Knowledge(PKLX):
    
    def parse(self, tokens: List[str], relations: List[str]) -> PKLX:
        try:
            return Binary().parse(tokens, relations)
        except ParsingException:
            pass
        try:
            return Unary().parse(tokens, relations)
        except ParsingException:
            pass
        raise ParsingException(f"Could not parse Knowledge: {tokens}")


class Binary(Knowledge):
    
    def __init__(self, left_expression=None, binary_operator=None, right_expression=None):
        self.left_expression = left_expression
        self.binary_operator = binary_operator
        self.right_expression = right_expression

    def parse(self, tokens: List[str], relations: List[str]) -> PKLX:
        if type(tokens) != list or len(tokens) == 0:
            raise ParsingException(f"Could not parse Binary: {tokens}")
        try:
            if tokens[0] == '(':
                # Find the matching closing parenthesis
                open_parentheses = 0
                for i, token in enumerate(tokens):
                    if token == '(':
                        open_parentheses += 1
                    if token == ')':
                        open_parentheses -= 1
                        if open_parentheses == 0:
                            break
                self.left_expression = NestedExpression().parse(tokens[0:i+1], relations)
                self.binary_operator = Binop().parse(tokens[i+1], relations)
                self.right_expression = Expression().parse(tokens[i+2:], relations)
            else:
                self.left_expression = Expression().parse([tokens[0]], relations)
                self.binary_operator = Binop().parse(tokens[1], relations)
                self.right_expression = Expression().parse(tokens[2:], relations)
            return self
        except ParsingException:
            raise ParsingException(f"Could not parse Binary: {tokens}")

    def to_graph(self) -> Tuple[str, nx.DiGraph]:
        left_expression_anchor, left_expression_graph = self.left_expression.to_graph()
        right_expression_anchor, right_expression_graph = self.right_expression.to_graph()
        binary_operator_anchor, binary_operator_graph = self.binary_operator.to_graph()
        graph = nx.compose_all([left_expression_graph, right_expression_graph, binary_operator_graph])
        graph.add_edge(left_expression_anchor, binary_operator_anchor)
        graph.add_edge(binary_operator_anchor, right_expression_anchor)
        return binary_operator_anchor, graph

    def contains(self, variable: str) -> bool:
        if self.left_expression.contains(variable):
            return True
        if self.right_expression.contains(variable):
            return True
        return False

    def __repr__(self):
        return f"{self.left_expression} {self.binary_operator} {self.right_expression}"


class Unary(Knowledge):
        
    def __init__(self, unary_operator=None, right_expression=None):
        self.unary_operator = unary_operator
        self.right_expression = right_expression

    def parse(self, tokens: List[str], relations: List[str]) -> PKLX:
        if type(tokens) != list or len(tokens) == 0:
            raise ParsingException(f"Could not parse Unary: {tokens}")
        try:
            self.unary_operator = Unop().parse(tokens[0], relations)
            self.right_expression = Expression().parse(tokens[1:], relations)
            return self
        except ParsingException:
            raise ParsingException(f"Could not parse Unary: {tokens}")

    def to_graph(self) -> Tuple[str, nx.DiGraph]:
        right_expression_anchor, right_expression_graph = self.right_expression.to_graph()
        unary_operator_anchor, unary_operator_graph = self.unary_operator.to_graph()
        graph = nx.compose(right_expression_graph, unary_operator_graph)
        graph.add_edge(unary_operator_anchor, right_expression_anchor)
        return unary_operator_anchor, graph

    def contains(self, variable: str) -> bool:
        if self.right_expression.contains(variable):
            return True
        return False

    def __repr__(self):
        return f"{self.unary_operator} {self.right_expression}"


class Expression():
    
    def parse(self, tokens: List[str], relations: List[str]) -> PKLX:
        if type(tokens) != list or len(tokens) == 0:
            raise ParsingException(f"Could not parse Expression: {tokens}")
        try:
            if len(tokens) == 1:
                return Name().parse(tokens[0], relations)
            else:
                return NestedExpression().parse(tokens, relations)
        except ParsingException:
            raise ParsingException(f"Could not parse Expression: {tokens}")


class NestedExpression(Expression):

    def __init__(self, knowledge=None):
        self.knowledge = knowledge

    def parse(self, tokens: List[str], relations: List[str]) -> PKLX:
        if type(tokens) != list or len(tokens) == 0:
            raise ParsingException(f"Could not parse NestedExpression: {tokens}")
        try:
            assert tokens[0] == '(' and tokens[-1] == ')'
            self.knowledge = Knowledge().parse(tokens[1:-1], relations)
            return self
        except (AssertionError, ParsingException):
            raise ParsingException(f"Could not parse NestedExpression: {tokens}")

    def to_graph(self) -> Tuple[str, nx.DiGraph]:
        return self.knowledge.to_graph()

    def contains(self, variable: str) -> bool:
        return self.knowledge.contains(variable)

    def __repr__(self):
        return f"( {self.knowledge} )"


class Name(Expression):
        
    def __init__(self, name=None):
        self.name = name

    def parse(self, token: str, relations: List[str]) -> PKLX:
        if type(token) == str and token[0].isalpha() and token not in relations:
            self.name = token
            return self
        else:
            raise ParsingException(f"Could not parse Name: {token}")

    def to_graph(self) -> Tuple[str, nx.DiGraph]:
        graph = nx.DiGraph()
        node = self.name
        graph.add_node(node, node_type='variable')
        return node, graph

    def contains(self, variable: str) -> bool:
        return self.name == variable
    
    def __repr__(self):
        return f"{self.name}"


class Binop():
    
    def __init__(self, name=None):
        self.name = name

    def parse(self, token: str, relations: List[str]) -> PKLX:
        if type(token) == str and token[0].isalpha() and token in relations:
            self.name = token
            return self
        else:
            raise ParsingException(f"Could not parse Binop: {token}")

    def to_graph(self) -> Tuple[str, nx.DiGraph]:
        graph = nx.DiGraph()
        node = str(uuid.uuid4()) + ' ' + self.name
        graph.add_node(node, node_type='binary')
        return node, graph

    def __repr__(self):
        return f"{self.name}"


class Unop():
        
    def __init__(self, name=None):
        self.name = name

    def parse(self, token: str, relations: List[str]) -> PKLX:
        if type(token) == str and token[0].isalpha() and token in relations:
            self.name = token
            return self
        else:
            raise ParsingException(f"Could not parse Unop: {token}")

    def to_graph(self) -> Tuple[str, nx.DiGraph]:
        graph = nx.DiGraph()
        node = str(uuid.uuid4()) + ' ' + self.name
        graph.add_node(node, node_type='unary')
        return node, graph

    def __repr__(self):
        return f"{self.name}"
